import os
import datetime
import numpy as np
from simtk import openmm as mm
from simtk import unit
from simtk.openmm.app.pdbfile import PDBFile
from simtk.openmm.app.topology import Topology
import simtk.openmm.app.element as elem
from cg_openmm.simulation.tools import build_mm_simulation
from cg_openmm.utilities.util import lj_v
from cg_openmm.utilities.iotools import write_pdbfile_without_topology

def add_new_elements(cgmodel):
        """
        Add coarse grained particle types to OpenMM.

        :param cgmodel: CGModel object (contains all attributes for a coarse grained model).
        :type cgmodel: class

        :returns: 
           - particle_list (list) - a list of the particles that were added to OpenMM's 'Element' List.

        :Example:

        >>> from foldamers.cg_model.cgmodel import CGModel
        >>> cgmodel = CGModel()
        >>> particle_types = add_new_elements(cgmodel)

        .. warning:: If the particle names were user defined, and any of the names conflict with existing element names in OpenMM, OpenMM will issue an error exit.

        """
        element_index = 117
        cg_particle_index = 1
        particle_list = []

        for monomer_type in cgmodel.monomer_types:
         for backbone_bead in range(monomer_type['backbone_length']):
          particle_name = str("X"+str(cg_particle_index))
          particle_symbol = str("X"+str(cg_particle_index))
          if particle_symbol not in elem.Element._elements_by_symbol:
           mass = cgmodel.get_particle_mass(cg_particle_index-1)
           elem.Element(element_index,particle_name,particle_symbol,mass)
           particle_list.append(particle_symbol)
           element_index = element_index + 1
          cg_particle_index = cg_particle_index + 1
          if type(monomer_type['sidechain_positions']) == int:
           sidechain_positions = [monomer_type['sidechain_positions']]
          else:
           sidechain_positions = monomer_type['sidechain_positions']
          if backbone_bead in sidechain_positions:
           for sidechain in range(monomer_type['sidechain_length']):
             particle_name = str("A"+str(cg_particle_index))
             particle_symbol = str("A"+str(cg_particle_index))
             if particle_symbol not in elem.Element._elements_by_symbol:
               mass = cgmodel.get_particle_mass(cg_particle_index-1)
               elem.Element(element_index,particle_name,particle_symbol,mass)
               particle_list.append(particle_symbol)
               element_index = element_index + 1
             cg_particle_index = cg_particle_index + 1
        return(particle_list)

def write_xml_file(cgmodel,xml_file_name):
        """
        Write an XML-formatted forcefield file for a coarse grained model.
         
        :param cgmodel: CGModel() class object.
        :type cgmodel: class

        :param xml_file_name: Path to XML output file.
        :type xml_file_name: str

        :Example:

        >>> from foldamers.cg_model.cgmodel import CGModel
        >>> cgmodel = CGModel()
        >>> xml_file_name = "openmm_cgmodel.xml"
        >>> write_xml_file(cgmodel,xml_file_name)

        """
        particle_list = add_new_elements(cgmodel)
        xml_object = open(xml_file_name,"w")
        xml_object.write("<ForceField>\n")
        xml_object.write(" <Info>\n")
        date = str(datetime.datetime.today()).split()[0]
        xml_object.write("  <DateGenerated> "+str(date)+" </DateGenerated>\n")
        xml_object.write("  <Source> https://github.com/shirtsgroup/cg_openmm </Source>\n")
        xml_object.write("  <Reference>\n")
        xml_object.write("  </Reference>\n")
        xml_object.write(" </Info>\n")
        xml_object.write(" <AtomTypes>\n")
        unique_particle_names = []
        unique_masses = []
        for particle_index in range(len(cgmodel.particle_list)):
          if cgmodel.particle_list[particle_index] not in unique_particle_names:
            unique_particle_names.append(cgmodel.particle_list[particle_index])
            unique_masses.append(cgmodel.get_particle_mass(particle_index))
        for particle_index in range(len(unique_particle_names)):          
          particle_type = cgmodel.get_particle_type(particle_index)
          xml_object.write('  <Type name="'+str(unique_particle_names[particle_index])+'" class="'+str(particle_type)+'" element="'+str(unique_particle_names[particle_index])+'" mass="'+str(unique_masses[particle_index]._value)+'"/>\n')
        xml_object.write(" </AtomTypes>\n")
        xml_object.write(" <Residues>\n")
        xml_object.write('  <Residue name="M">\n')
        for particle_index in range(len(unique_particle_names)):
          xml_object.write('   <Atom name="'+str(unique_particle_names[particle_index])+'" type="'+str(unique_particle_names[particle_index])+'"/>\n')
        for bond in cgmodel.bond_list:
          if all(bond[i] < len(unique_particle_names) for i in range(2)): 
            particle_1_name = cgmodel.get_particle_name(bond[0])
            particle_2_name = cgmodel.get_particle_name(bond[1])
            xml_object.write('   <Bond atomName1="'+str(particle_1_name)+'" atomName2="'+str(particle_2_name)+'"/>\n')
        xml_object.write('   <ExternalBond atomName="'+str(unique_particle_names[0])+'"/>\n')
        external_parent = unique_particle_names[len(unique_particle_names)-cgmodel.monomer_types[0]['sidechain_length']-1]
        xml_object.write('   <ExternalBond atomName="'+str(external_parent)+'"/>\n')
        xml_object.write("  </Residue>\n")
        xml_object.write('  <Residue name="MT">\n')
        for particle_index in range(len(unique_particle_names)):
          xml_object.write('   <Atom name="'+str(unique_particle_names[particle_index])+'" type="'+str(unique_particle_names[particle_index])+'"/>\n')
        for bond in cgmodel.bond_list:
          if all(bond[i] < len(unique_particle_names) for i in range(2)):
            particle_1_name = cgmodel.get_particle_name(bond[0])
            particle_2_name = cgmodel.get_particle_name(bond[1])
            xml_object.write('   <Bond atomName1="'+str(particle_1_name)+'" atomName2="'+str(particle_2_name)+'"/>\n')
        xml_object.write('   <ExternalBond atomName="'+str(external_parent)+'"/>\n')
        xml_object.write("  </Residue>\n")
        xml_object.write(" </Residues>\n")
        if cgmodel.include_bond_forces:
          xml_object.write(" <HarmonicBondForce>\n")
          unique_bond_list = []
          for bond in cgmodel.bond_list:
            if any(bond[i] < len(unique_particle_names) for i in range(2)):
              unique_bond_list.append(bond)
          for bond in unique_bond_list:
            particle_1_name = cgmodel.get_particle_name(bond[0])
            particle_2_name = cgmodel.get_particle_name(bond[1])
            xml_type_1 = particle_1_name #unique_particle_names.index(particle_1_name)
            xml_type_2 = particle_2_name #unique_particle_names.index(particle_2_name)
            bond_length = cgmodel.get_bond_length(bond[0],bond[1]).in_units_of(unit.nanometer)._value
            bond_force_constant = cgmodel.get_bond_force_constant(bond[0],bond[1])
            xml_object.write('  <Bond type1="'+str(xml_type_1)+'" type2="'+str(xml_type_2)+'" length="'+str(bond_length)+'" k="'+str(bond_force_constant)+'"/>\n')
          xml_object.write(" </HarmonicBondForce>\n")
        if cgmodel.include_bond_angle_forces:
          xml_object.write(" <HarmonicAngleForce>\n")
          unique_angle_list = []
          for angle in cgmodel.bond_angle_list:
            if any(angle[i] < len(unique_particle_names) for i in range(3)):
              unique_angle_list.append(angle)
          for angle in unique_angle_list:
            bond_angle_force_constant = cgmodel.get_bond_angle_force_constant(angle[0],angle[1],angle[2])
            equil_bond_angle = cgmodel.get_equil_bond_angle(angle[0],angle[1],angle[2])
            particle_1_name = cgmodel.get_particle_name(angle[0])
            particle_2_name = cgmodel.get_particle_name(angle[1])
            particle_3_name = cgmodel.get_particle_name(angle[2])
            xml_type_1 = particle_1_name #unique_particle_names.index(particle_1_name)
            xml_type_2 = particle_2_name #unique_particle_names.index(particle_2_name)
            xml_type_3 = particle_3_name #unique_particle_names.index(particle_3_name)
            xml_object.write('  <Angle angle="'+str(equil_bond_angle)+'" k="'+str(bond_angle_force_constant)+'" type1="'+str(xml_type_1)+'" type2="'+str(xml_type_2)+'" type3="'+str(xml_type_3)+'"/>\n')
          xml_object.write(" </HarmonicAngleForce>\n")
        if cgmodel.include_torsion_forces:
          xml_object.write(' <PeriodicTorsionForce ordering="amber">\n')
          unique_torsion_list = []
          #print(cgmodel.torsion_list)
          for torsion in cgmodel.torsion_list:
            if any(torsion[i] < len(unique_particle_names) for i in range(4)):
              unique_torsion_list.append(torsion)
          for torsion in unique_torsion_list:
            torsion_force_constant = cgmodel.get_torsion_force_constant([torsion[0],torsion[1],torsion[2],torsion[3]])
            equil_torsion_angle = cgmodel.get_equil_torsion_angle([torsion[0],torsion[1],torsion[2],torsion[3]])
            particle_1_name = cgmodel.get_particle_name(torsion[0])
            particle_2_name = cgmodel.get_particle_name(torsion[1])
            particle_3_name = cgmodel.get_particle_name(torsion[2])
            particle_4_name = cgmodel.get_particle_name(torsion[3])
            xml_type_1 = particle_1_name #unique_particle_names.index(particle_1_name)
            xml_type_2 = particle_2_name #unique_particle_names.index(particle_2_name)
            xml_type_3 = particle_3_name #unique_particle_names.index(particle_3_name)
            xml_type_4 = particle_4_name #unique_particle_names.index(particle_4_name)
            periodicity = cgmodel.get_torsion_periodicity(torsion)
            xml_object.write('  <Proper k1="'+str(torsion_force_constant)+'" periodicity1="'+str(periodicity)+'" phase1="'+str(equil_torsion_angle)+'" type1="'+str(xml_type_1)+'" type2="'+str(xml_type_2)+'" type3="'+str(xml_type_3)+'" type4="'+str(xml_type_4)+'"/>\n')
          xml_object.write(" </PeriodicTorsionForce>\n")
        if cgmodel.include_nonbonded_forces:
          xml_object.write(' <NonbondedForce coulomb14scale="0.833333" lj14scale="0.5">\n')
          for particle_index in range(len(unique_particle_names)):
            charge = cgmodel.get_particle_charge(particle_index)._value
            sigma = cgmodel.get_sigma(particle_index).in_units_of(unit.nanometer)._value
            epsilon = cgmodel.get_epsilon(particle_index)._value  
            particle_name = cgmodel.get_particle_name(particle_index)
            xml_object.write('  <Atom type="'+str(particle_name)+'" charge="'+str(charge)+'" sigma="'+str(sigma)+'" epsilon="'+str(epsilon)+'"/>\n')
          xml_object.write(" </NonbondedForce>\n")
        xml_object.write("</ForceField>\n")
        xml_object.close()
        return

def verify_topology(cgmodel):
        """
        Given a coarse grained model that contains a Topology() (cgmodel.topology), this function verifies the validity of the topology.

        :param cgmodel: CGModel() class object.
        :type cgmodel: class

        :Example:

        >>> from foldamers.cg_model.cgmodel import CGModel
        >>> cgmodel = CGModel()
        >>> verify_topology(cgmodel)

        .. warning:: The function will force an error exit if the topology is invalid, and will proceed as normal if the topology is valid. 

        """
        if cgmodel.num_beads != cgmodel.topology.getNumAtoms():
          print("ERROR: The number of particles in the coarse grained model\n")
          print("does not match the number of particles in the OpenMM topology.\n")
          print("There are "+str(cgmodel.num_beads)+" particles in the coarse grained model\n")
          print("and "+str(cgmodel.topology.getNumAtoms())+" particles in the OpenMM topology.")
          exit()

        if cgmodel.polymer_length != cgmodel.topology.getNumResidues():
          print("ERROR: The number of monomers in the coarse grained model\n")
          print("does not match the number of residues in the OpenMM topology.\n")
          print("There are "+str(cgmodel.polymer_length)+" monomers in the coarse grained model\n")
          print("and "+str(cgmodel.topology.getNumResidues())+" monomers in the OpenMM topology.")
          exit()

        return

def build_topology(cgmodel,use_pdbfile=False,pdbfile=None):
        """

        Construct an OpenMM `Topology() <https://simtk.org/api_docs/openmm/api4_1/python/classsimtk_1_1openmm_1_1app_1_1topology_1_1Topology.html>`_ class object for our coarse grained model,

        :param cgmodel: CGModel() class object
        :type cgmodel: class

        :param use_pdbfile: Determines whether or not to use a PDB file in order to generate the Topology().
        :type use_pdbfile: Logical

        :param pdbfile: Name of a PDB file to use when building the topology.
        :type pdbfile: str

        :returns:
            - topology (`Topology() <https://simtk.org/api_docs/openmm/api4_1/python/classsimtk_1_1openmm_1_1app_1_1topology_1_1Topology.html>`_ ) - OpenMM Topology() object

        :Example:

        >>> from foldamers.cg_model.cgmodel import CGModel
        >>> from foldamers.util.iotools import write_pdbfile_without_topology
        >>> input_pdb = "top.pdb"
        >>> cgmodel = CGModel()
        >>> write_pdbfile_without_topology(cgmodel,input_pdb)
        >>> topology = build_topology(cgmodel,use_pdbfile=True,pdbfile=input_pdb)
        >>> cgmodel.topology = topology

        .. warning:: When 'use_pdbfile'=True, this function will use the `PDBFile() <https://simtk.org/api_docs/openmm/api4_1/python/classsimtk_1_1openmm_1_1app_1_1pdbfile_1_1PDBFile.html>`_ class object from OpenMM to build the Topology().  In order for this approach to function correctly, the particle names in the PDB file must match the particle names in the coarse grained model.

        """
        if cgmodel.constrain_bonds == True:
         use_pdbfile = True
        if use_pdbfile == True:
         if pdbfile == None:
           write_pdbfile_without_topology(cgmodel,"topology_source.pdb")
           pdb = PDBFile("topology_source.pdb")
           topology = pdb.getTopology()
           os.remove("topology_source.pdb")
           return(topology)
         else:
           pdb = PDBFile(pdbfile)
           topology = pdb.getTopology()
           return(topology)

        topology = Topology()

        chain = topology.addChain()
        residue_index = 1
        cg_particle_index = 0
        for monomer_type in cgmodel.sequence:
         residue = topology.addResidue(str(residue_index), chain)
         for backbone_bead in range(monomer_type['backbone_length']):
          particle_symbol = cgmodel.get_particle_name(cg_particle_index)
          element = elem.Element.getBySymbol(particle_symbol)
          particle = topology.addAtom(particle_symbol, element, residue)
          if backbone_bead == 0 and residue_index != 1:
           if cgmodel.include_bond_forces or cgmodel.constrain_bonds:
            topology.addBond(particle,last_backbone_particle)
          last_backbone_particle = particle
          cg_particle_index = cg_particle_index + 1
          if backbone_bead in [monomer_type['sidechain_positions']]:
           for sidechain_bead in range(monomer_type['sidechain_length']):
             particle_symbol = cgmodel.get_particle_name(cg_particle_index)
             element = elem.Element.getBySymbol(particle_symbol)
             particle = topology.addAtom(particle_symbol, element, residue)
             if sidechain_bead == 0:
              if cgmodel.include_bond_forces or cgmodel.constrain_bonds:
               topology.addBond(particle,last_backbone_particle)
             if sidechain_bead != 0:
              if cgmodel.include_bond_forces or cgmodel.constrain_bonds:
               topology.addBond(particle,last_sidechain_particle)
             last_sidechain_particle = particle
             cg_particle_index = cg_particle_index + 1
         residue_index = residue_index + 1
        cgmodel.topology = topology
        verify_topology(cgmodel)
        return(topology)


def get_num_forces(cgmodel):
        """
        Given a CGModel() class object, this function determines how many forces we are including when evaluating the energy.

        :param cgmodel: CGModel() class object
        :type cgmodel: class

        :returns: 
            - total_forces (int) - Number of forces in the coarse grained model

        :Example:

        >>> from foldamers.cg_model.cgmodel import CGModel
        >>> cgmodel = CGModel()
        >>> total_number_forces = get_num_forces(cgmodel)

        """
        total_forces = 0
        if cgmodel.include_bond_forces: total_forces = total_forces + 1
        if cgmodel.include_nonbonded_forces: total_forces = total_forces + 1
        if cgmodel.include_bond_angle_forces: total_forces = total_forces + 1
        if cgmodel.include_torsion_forces: total_forces = total_forces + 1
        return(total_forces)

def verify_system(cgmodel):
        """
        Given a CGModel() class object, this function confirms that its OpenMM `System() <https://simtk.org/api_docs/openmm/api4_1/python/classsimtk_1_1openmm_1_1openmm_1_1System.html>`_ object is configured correctly.

        :param cgmodel: CGModel() class object
        :type cgmodel: class

        :Example:

        >>> from foldamers.cg_model.cgmodel import CGModel
        >>> cgmodel = CGModel()
        >>> verify_system(cgmodel)

        .. warning:: The function will force an error exit if the system is invalid, and will proceed as normal if the system is valid.

        """

        if get_num_forces(cgmodel) != cgmodel.system.getNumForces():
          print("ERROR: the number of forces included in the coarse grained model\n")
          print("does not match the number of forces in the OpenMM system object.\n")
          print(" There are "+str(get_num_forces(cgmodel))+" forces in the coarse grained model\n")
          print("and "+str(cgmodel.system.getNumForces())+" forces in the OpenMM System().")
          exit()

        if cgmodel.num_beads != cgmodel.system.getNumParticles():
          print("ERROR: The number of particles in the coarse grained model\n")
          print("does not match the number of particles in the OpenMM system.\n")
          print("There are "+str(cgmodel.num_beads)+" particles in the coarse grained model\n")
          print("and "+str(cgmodel.ssytem.getNumParticles())+" particles in the OpenMM system.")
          exit()

        if cgmodel.constrain_bonds:
          if len(cgmodel.bond_list) != cgmodel.system.getNumConstraints():
            print("ERROR: Bond constraints were requested, but the\n")
            print("number of constraints in the coarse grained model\n")
            print("does not match the number of constraintes in the OpenMM system.\n")
            print("There are "+str(cgmodel.bond_list)+" bond constraints in the coarse grained model\n")
            print("and "+str(cgmodel.system.getNumConstraints())+" constraints in the OpenMM system.")
            exit()

        return

def test_force(cgmodel,force,force_type=None):
        """

        Given an OpenMM `Force() <https://simtk.org/api_docs/openmm/api4_1/python/classsimtk_1_1openmm_1_1openmm_1_1Force.html>`_, this function determines if there are any problems with its configuration.

        :param cgmodel: CGModel() class object.
        :type cgmodel: class

        :param force: An OpenMM Force() object.
        :type force: `Force() <https://simtk.org/api_docs/openmm/api4_1/python/classsimtk_1_1openmm_1_1openmm_1_1Force.html>`_, this function determines if there are any problems with its configuration.

        :param force_type: Designates the kind of 'force' provided. (Valid options include: "Nonbonded")
        :type force_type: str

        :returns: 
            - 'success' (Logical) - a variable indicating if the force test passed.

        :Example:

        >>> from simtk.openmm.openmm import NonbondedForce
        >>> from foldamers.cg_model.cgmodel import CGModel
        >>> cgmodel = CGModel()
        >>> force = NonbondedForce()
        >>> force_type = "Nonbonded"
        >>> test_result = test_force(cgmodel,force,force_type="Nonbonded")

        """
        success=True
        if force_type == "Nonbonded":
           if cgmodel.num_beads != force.getNumParticles():
             print("ERROR: The number of particles in the coarse grained model is different")
             print("from the number of particles with nonbonded force definitions in the OpenMM NonbondedForce.\n")
             print("There are "+str(cgmodel.num_beads)+" particles in the coarse grained model")
             print("and "+str(force.getNumParticles())+" particles in the OpenMM NonbondedForce.")
             success=False

           total_nonbonded_energy = 0.0 * unit.kilojoule_per_mole
           #print(cgmodel.nonbonded_interaction_list)
           for nonbonded_interaction in cgmodel.nonbonded_interaction_list:
             particle_1_positions = cgmodel.positions[nonbonded_interaction[0]]
             particle_2_positions = cgmodel.positions[nonbonded_interaction[1]]
             sigma = cgmodel.get_sigma(nonbonded_interaction[0])
             epsilon = cgmodel.get_epsilon(nonbonded_interaction[0])
             int_energy = lj_v(particle_1_positions,particle_2_positions,sigma,epsilon)
             total_nonbonded_energy = total_nonbonded_energy.__add__(int_energy)
           
           cgmodel.include_bond_forces = False
           cgmodel.include_bond_angle_forces = False
           cgmodel.include_torsion_forces = False
           cgmodel.topology = build_topology(cgmodel)
           cgmodel.simulation = build_mm_simulation(cgmodel.topology,cgmodel.system,cgmodel.positions,simulation_time_step=5.0*unit.femtosecond,print_frequency=1)
           potential_energy = cgmodel.simulation.context.getState(getEnergy=True).getPotentialEnergy()

           #if potential_energy.__sub__(total_nonbonded_energy).__gt__(0.1 * unit.kilojoule_per_mole):
             #print("Warning: The nonbonded potential energy computed by hand does not agree")
             #print("with the value computed by OpenMM.")
             #print("The value computed by OpenMM was: "+str(potential_energy))
             #print("The value computed by hand was: "+str(total_nonbonded_energy))
             #print("Check the units for your model parameters.  If the problem persists, there")
             #print("could be some other problem with the configuration of your coarse grained model.")
             #success = False
           #else:
             #print("The OpenMM nonbonded energy matches the energy computed by hand:")
             #print(str(potential_energy))

        return(success)

def add_rosetta_exception_parameters(cgmodel,nonbonded_force,particle_index_1,particle_index_2):
        """
        """
        exception_list = []
        for exception in range(nonbonded_force.getNumExceptions()):
          index_1,index_2,charge,sigma,epsilon = nonbonded_force.getExceptionParameters(exception)
          if [index_1,index_2] not in exception_list and [index_2,index_1] not in exception_list:
            exception_list.append([index_1,index_2])
        
        if [particle_index_1,particle_index_2] not in exception_list and [particle_index_2,particle_index_1] not in exception_list:
          charge_1 = cgmodel.get_particle_charge(particle_index_1)
          sigma_1 = cgmodel.get_sigma(particle_index_1).in_units_of(unit.nanometer)
          epsilon_1 = cgmodel.get_epsilon(particle_index_1).in_units_of(unit.kilojoule_per_mole)
          charge_2 = cgmodel.get_particle_charge(particle_index_2)
          sigma_2 = cgmodel.get_sigma(particle_index_2).in_units_of(unit.nanometer)
          epsilon_2 = cgmodel.get_epsilon(particle_index_2).in_units_of(unit.kilojoule_per_mole)
          nonbonded_force.addException(particle_index_1,particle_index_2,0.2*charge_1*charge_2,0.2*sigma_1*sigma_2,0.2*epsilon_1*epsilon_2)
        return(nonbonded_force)

def add_force(cgmodel,force_type=None,rosetta_scoring=False):
        """

        Given a 'cgmodel' and 'force_type' as input, this function adds
        the OpenMM force corresponding to 'force_type' to 'cgmodel.system'.

        :param cgmodel: CGModel() class object.
        :param type: class

        :param force_type: Designates the kind of 'force' provided. (Valid options include: "Bond", "Nonbonded", "Angle", and "Torsion")
        :type force_type: str

        :returns: 
             - cgmodel (class) - 'foldamers' CGModel() class object
             - force (class) - An OpenMM `Force() <https://simtk.org/api_docs/openmm/api4_1/python/classsimtk_1_1openmm_1_1openmm_1_1Force.html>`_ object.

        :Example:

        >>> from foldamers.cg_model.cgmodel import CGModel
        >>> cgmodel = CGModel()
        >>> force_type = "Bond"
        >>> cgmodel,force = add_force(cgmodel,force_type=force_type)

        """
        if force_type == "Bond":

          bond_force = mm.HarmonicBondForce()
          bond_list = []

          for bond_indices in cgmodel.get_bond_list():
              bond_list.append([bond_indices[0],bond_indices[1]])
              if cgmodel.include_bond_forces:
                bond_force_constant = cgmodel.get_bond_force_constant(bond_indices[0],bond_indices[1])
                bond_length = cgmodel.get_bond_length(bond_indices[0],bond_indices[1]).in_units_of(unit.nanometer)._value
                bond_force.addBond(bond_indices[0],bond_indices[1],bond_length,bond_force_constant)
              if cgmodel.constrain_bonds:
                bond_length = cgmodel.get_bond_length(bond_indices[0],bond_indices[1]).in_units_of(unit.nanometer)._value
                if not cgmodel.include_bond_forces:
                  bond_force_constant = 0.0 * cgmodel.get_bond_force_constant(bond_indices[0],bond_indices[1])
                  bond_force.addBond(bond_indices[0],bond_indices[1],bond_length,bond_force_constant)
                cgmodel.system.addConstraint(bond_indices[0],bond_indices[1],bond_length)

          if len(bond_list) != bond_force.getNumBonds():
            print("ERROR: The number of bonds in the coarse grained model is different\n")
            print("from the number of bonds in its OpenMM System object\n")
            print("There are "+str(len(bond_list))+" bonds in the coarse grained model\n")
            print("and "+str(bond_force.getNumBonds())+" bonds in the OpenMM system object.")
            exit()

          cgmodel.system.addForce(bond_force)
          force = bond_force

        if force_type == "Nonbonded":

          nonbonded_force = mm.NonbondedForce()
          nonbonded_force.setNonbondedMethod(mm.NonbondedForce.NoCutoff)

          for particle in range(cgmodel.num_beads):
            charge = cgmodel.get_particle_charge(particle)
            sigma = cgmodel.get_sigma(particle)
            epsilon = cgmodel.get_epsilon(particle)
            nonbonded_force.addParticle(charge,sigma,epsilon)

          if len(cgmodel.bond_list) >= 1:
            if not rosetta_scoring:
              nonbonded_force.createExceptionsFromBonds(cgmodel.bond_list,1.0,1.0)
            if rosetta_scoring:
              nonbonded_force.createExceptionsFromBonds(cgmodel.bond_list,0.0,0.0)
              for torsion in cgmodel.torsion_list:
                for bond in cgmodel.bond_list:
                  if bond[0] not in torsion:
                    if bond[1] == torsion[0]:
                      nonbonded_force = add_rosetta_exception_parameters(cgmodel,nonbonded_force,bond[0],torsion[3])
                    if bond[1] == torsion[3]:
                      nonbonded_force = add_rosetta_exception_parameters(cgmodel,nonbonded_force,bond[0],torsion[0])
                  if bond[1] not in torsion:
                    if bond[0] == torsion[0]:
                      nonbonded_force = add_rosetta_exception_parameters(cgmodel,nonbonded_force,bond[1],torsion[3])
                    if bond[0] == torsion[3]:
                      nonbonded_force = add_rosetta_exception_parameters(cgmodel,nonbonded_force,bond[1],torsion[0])
          cgmodel.system.addForce(nonbonded_force)
          force = nonbonded_force
          #for particle in range(cgmodel.num_beads):
            #print(force.getParticleParameters(particle))

        if force_type == "Angle":
          angle_force = mm.HarmonicAngleForce()  
          for angle in cgmodel.bond_angle_list:
            bond_angle_force_constant = cgmodel.get_bond_angle_force_constant(angle[0],angle[1],angle[2])
            equil_bond_angle = cgmodel.get_equil_bond_angle(angle[0],angle[1],angle[2])
            angle_force.addAngle(angle[0],angle[1],angle[2],equil_bond_angle,bond_angle_force_constant)
          cgmodel.system.addForce(angle_force)
          force = angle_force

        if force_type == "Torsion":
          torsion_force = mm.PeriodicTorsionForce()
          for torsion in cgmodel.torsion_list:
            torsion_force_constant = cgmodel.get_torsion_force_constant(torsion)
            equil_torsion_angle = cgmodel.get_equil_torsion_angle(torsion)
            periodicity = cgmodel.get_torsion_periodicity(torsion)
            torsion_force.addTorsion(torsion[0],torsion[1],torsion[2],torsion[3],periodicity,equil_torsion_angle,torsion_force_constant)
              #print(torsion_force.getNumTorsions())
          cgmodel.system.addForce(torsion_force)
          force = torsion_force

        return(cgmodel,force)

def test_forces(cgmodel):
        """
        Given a cgmodel that contains positions and an
        an OpenMM System() object, this function tests 
        the forces for cgmodel.system.

        More specifically, this function confirms that the
        model does not have any "NaN" or unphysically large forces.

        :param cgmodel: CGModel() class object.
        :param type: class

        :returns:
            - success (Logical) - Indicates if this cgmodel has unphysical forces.

        :Example:

        >>> from foldamers.cg_model.cgmodel import CGModel
        >>> cgmodel = CGModel()
        >>> pass_forces_test = test_forces(cgmodel)

        """
        if cgmodel.topology == None:
          cgmodel.topology = build_topology(cgmodel)
        simulation = build_mm_simulation(cgmodel.topology,cgmodel.system,cgmodel.positions,simulation_time_step=5.0*unit.femtosecond,print_frequency=1)
        forces = simulation.context.getState(getForces=True).getForces()
        success = True
        for force in forces:
          for component in force:
            if 'nan' in str(component):
              print("Detected 'nan' force value")
              print("for particle "+str(forces.index(force)))
              success = False
              return(success)
            if component.__gt__(9.9e9 * component.unit):
              print("Detected unusually large forces")
              print("for particle "+str(forces.index(force)))
              print("The force is: "+str("{:.2e}".format(component._value))+" "+str(component.unit))
              success = False
              return(success)
        return(success)

def build_system(cgmodel,rosetta_scoring=False,verify=True):
        """
        Builds an OpenMM `System() <https://simtk.org/api_docs/openmm/api4_1/python/classsimtk_1_1openmm_1_1openmm_1_1System.html>`_ object, given a CGModel() as input.

        :param cgmodel: CGModel() class object
        :type cgmodel: class

        :returns: 
            - system ( `System() <https://simtk.org/api_docs/openmm/api4_1/python/classsimtk_1_1openmm_1_1openmm_1_1System.html>`_ ) - OpenMM System() object

        :Example:

        >>> from foldamers.cg_model.cgmodel import CGModel
        >>> cgmodel = CGModel()
        >>> system = build_system(cgmodel)
        >>> cgmodel.system = system

        """
        # Create system
        system = mm.System()
        for particle in range(cgmodel.num_beads):
            mass = cgmodel.get_particle_mass(particle)
            system.addParticle(mass)
        cgmodel.system = system

        #length_scale = cgmodel.bond_lengths['bb_bb_bond_length']
        #box_vectors = [[100.0*length_scale._value,0.0,0.0],[0.0,100.0*length_scale._value,0.0],[0.0,0.0,100.0*length_scale._value]]
        #system.setDefaultPeriodicBoxVectors(box_vectors[0],box_vectors[1],box_vectors[2])

        if cgmodel.include_bond_forces or cgmodel.constrain_bonds:
         # Create bond (harmonic) potentials
         cgmodel,bond_force = add_force(cgmodel,force_type="Bond")
         if cgmodel.positions != None:
          if not test_force(cgmodel,bond_force,force_type="Bond"):
           print("ERROR: The bond force definition is giving 'nan'")
           exit()

        if cgmodel.include_nonbonded_forces:
         # Create nonbonded forces
          cgmodel,nonbonded_force = add_force(cgmodel,force_type="Nonbonded",rosetta_scoring=rosetta_scoring)

          #if cgmodel.positions != None:
           #print("Testing the nonbonded forces")
           #if not test_force(cgmodel,nonbonded_force,force_type="Nonbonded"):
            #print("ERROR: there was a problem with the nonbonded force definitions.")
            #exit()

        if cgmodel.include_bond_angle_forces:
          # Create bond angle potentials
          cgmodel,bond_angle_force = add_force(cgmodel,force_type="Angle")
          if cgmodel.positions != None:
            if not test_force(cgmodel,bond_angle_force,force_type="Angle"):
              print("ERROR: There was a problem with the bond angle force definitions.")
              exit()

        if cgmodel.include_torsion_forces:
          # Create torsion potentials
          cgmodel,torsion_force = add_force(cgmodel,force_type="Torsion")
          if cgmodel.positions != None:
            if not test_force(cgmodel,torsion_force,force_type="Torsion"):
              print("ERROR: There was a problem with the torsion definitions.")
              exit()

        if verify:
          if cgmodel.positions != None:
            if not test_forces(cgmodel):
              print("ERROR: There was a problem with the forces.")
              exit()

        return(system)



import numpy as np
from simtk import openmm as mm
from simtk.openmm import *
from simtk import unit
import simtk.openmm.app.element as elem
from simtk.openmm.app import *


def distance(positions_1,positions_2):
        """
        Calculate the distance between two particles, given their positions.

        :param positions_1: Positions for the first particle
        :type positions_1: `Quantity() <http://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_ ( np.array( [3] ), simtk.unit )

        :param positions_2: Positions for the first particle
        :type positions_2: `Quantity() <http://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_ ( np.array( [3] ), simtk.unit )

        :returns:
            - distance ( `Quantity() <http://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_) - Distance between two particles

        :Example:

        >>> from foldamers.cg_model.cgmodel import CGModel
        >>> cgmodel = CGModel()
        >>> particle_1_coordinates = cgmodel.positions[0]
        >>> particle_2_coordinates = cgmodel.positions[1]
        >>> particle_distance = distance(particle_1_coordinates,particle_2_coordinates)
  
        """

        direction_comp = np.zeros(3) * positions_1.unit

        for direction in range(len(direction_comp)):
          direction_comp[direction] = positions_1[direction].__sub__(positions_2[direction])

        direction_comb = np.zeros(3) * positions_1.unit.__pow__(2.0)
        for direction in range(3):
          direction_comb[direction] = direction_comp[direction].__pow__(2.0)

        sqrt_arg = direction_comb[0].__add__(direction_comb[1]).__add__(direction_comb[2])

        value = math.sqrt(sqrt_arg._value)
        units = sqrt_arg.unit.sqrt()
        distance = unit.Quantity(value=value,unit=units)

        return(distance)

def get_box_vectors(box_size):
        """
        Given a simulation box length, construct a vector.

        :param box_size: Length of individual sides of a simulation box
        :type box_size: `Quantity() <http://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_ ( float, simtk.unit )

        :returns: 
             - box_vectors ( List( `Quantity() <http://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_ ) ) - Vectors to use when defining an OpenMM simulation box.

        """

        units = box_size.unit
        a = unit.Quantity(np.zeros([3]), units)
        a[0] = box_size
        b = unit.Quantity(np.zeros([3]), units)
        b[1] = box_size
        c = unit.Quantity(np.zeros([3]), units)
        c[2] = box_size
        box_vectors = [a,b,c]
        return(box_vectors)

def set_box_vectors(system,box_size):
        """
        Impose a set of simulation box vectors on an OpenMM simulation object.

        :param system: OpenMM System()
        :type system: `System() <https://simtk.org/api_docs/openmm/api4_1/python/classsimtk_1_1openmm_1_1openmm_1_1System.html>`_

        :param box_size: Length of individual sides of a simulation box
        :type box_size: `Quantity() <http://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_ ( float, simtk.unit )

        :returns: 
            - system (`System() <https://simtk.org/api_docs/openmm/api4_1/python/classsimtk_1_1openmm_1_1openmm_1_1System.html>`_) - OpenMM system object

        """

        a,b,c = get_box_vectors(box_size)
        system.setDefaultPeriodicBoxVectors(a, b, c)
        return(system)

def lj_v(positions_1,positions_2,sigma,epsilon):
        """
        Calculate the Lennard-Jones interaction energy between two particles, given their positions and definitions for their equilbrium interaction distance (sigma) and strength (epsilon).

        :param positions_1: Positions for the first particle
        :type positions_1: `Quantity() <http://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_ ( np.array( [3] ), simtk.unit )

        :param positions_2: Positions for the first particle
        :type positions_2: `Quantity() <http://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_ ( np.array( [3] ), simtk.unit )

        :param sigma: Lennard-Jones equilibrium interaction distance for two non-bonded particles
        :type sigma: `Quantity() <http://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_

        :param epsilon: Lennard-Jones equilibrium interaction energy for two non-bonded particles.
        :type epsilon: `Quantity() <http://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_

        :returns:
           - v ( `Quantity() <http://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_ ) - Lennard-Jones interaction energy

        """

        dist = distance(positions_1,positions_2)
        quot = sigma.__div__(dist)
        attr = quot.__pow__(6.0)
        rep = quot.__pow__(12.0)
        v = 4.0*epsilon.__mul__(rep.__sub__(attr))
        return(v)




import os
import subprocess
import numpy as np
import matplotlib.pyplot as pyplot
from simtk import unit
import openmmtools as mmtools
from cg_openmm.utilities.util import set_box_vectors, get_box_vectors
from cg_openmm.simulation.tools import get_simulation_time_step
from simtk.openmm.app.pdbfile import PDBFile
from mdtraj.formats import PDBTrajectoryFile
from mdtraj import Topology
from yank import mpi, analyze
from yank.analyze import extract_trajectory
from yank.multistate import MultiStateReporter, MultiStateSampler, ReplicaExchangeSampler
from yank.multistate import ReplicaExchangeAnalyzer
from yank.utils import config_root_logger
# quiet down some citation spam
MultiStateSampler._global_citation_silence = True

kB = 0.008314462

def make_replica_pdb_files(topology,replica_positions):
        """
        Make PDB files from replica exchange simulation trajectory data

        :param topology: OpenMM Topology
        :type topology: `Topology() <https://simtk.org/api_docs/openmm/api4_1/python/classsimtk_1_1openmm_1_1app_1_1topology_1_1Topology.html>`_

        :param replica_positions: Positions array for the replica exchange data for which we will write PDB files
        :type replica_positions: `Quantity() <http://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_ ( np.array( [n_replicas,cgmodel.num_beads,3] ), simtk.unit )

        :returns:
            - file_list ( List( str ) ) - A list of names for the files that were written

        :Example:

        >>> from foldamers.cg_model.cgmodel import CGModel
        >>> from cg_openmm.simulation.rep_exch import *
        >>> cgmodel = CGModel()
        >>> replica_energies,replica_positions,replica_state_indices = run_replica_exchange(cgmodel.topology,cgmodel.system,cgmodel.positions)
        >>> pdb_file_list = make_replica_pdb_files(cgmodel.topology,replica_positions)

        """
        replica_index = 1
        file_list = []
        for replica_index in range(len(replica_positions)):
          replica_trajectory = replica_positions[replica_index]
          file_name = str("replica_"+str(replica_index+1)+".pdb")
          file = open(file_name,"w")
          PDBFile.writeHeader(topology,file=file)
          modelIndex=1
          for positions in replica_trajectory:
            PDBFile.writeModel(topology,positions,file=file,modelIndex=modelIndex)
          PDBFile.writeFooter(topology,file=file)
          file.close()
          file_list.append(file_name)
        return(file_list)

def read_replica_exchange_data(system=None,topology=None,temperature_list=None,output_data="output.nc",print_frequency=None):
        """
        Read replica exchange simulation data.

        :param system: OpenMM system object, default = None
        :type system: `System() <https://simtk.org/api_docs/openmm/api4_1/python/classsimtk_1_1openmm_1_1openmm_1_1System.html>`_

        :param topology: OpenMM topology object, default = None
        :type topology: `Topology() <https://simtk.org/api_docs/openmm/api4_1/python/classsimtk_1_1openmm_1_1app_1_1topology_1_1Topology.html>`_

        :param temperature_list: List of temperatures that will be used to define different replicas (thermodynamics states), default = None
        :type temperature_list: List( `SIMTK <https://simtk.org/>`_ `Unit() <http://docs.openmm.org/7.1.0/api-python/generated/simtk.unit.unit.Unit.html>`_ * number_replicas ) 

        :param output_data: Path to the output data for a Yank, NetCDF-formatted file containing replica exchange simulation data, default = None
        :type output_data: str

        :param print_frequency: Number of simulation steps to skip when writing data, default = None
        :type print_frequency: int

        :returns: 
            - replica_energies ( `Quantity() <http://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_ ( np.float( [number_replicas,number_simulation_steps] ), simtk.unit ) ) - The potential energies for all replicas at all (printed) time steps
            - replica_positions ( `Quantity() <http://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_ ( np.float( [number_replicas,number_simulation_steps,cgmodel.num_beads,3] ), simtk.unit ) ) - The positions for all replicas at all (printed) time steps
 
            - replica_state_indices ( np.int64( [number_replicas,number_simulation_steps] ), simtk.unit ) - The thermodynamic state assignments for all replicas at all (printed) time steps

        :Example:

        >>> from foldamers.cg_model.cgmodel import CGModel
        >>> from cg_openmm.simulation.rep_exch import *
        >>> cgmodel = CGModel()
        >>> replica_energies,replica_positions,replica_state_indices = run_replica_exchange(cgmodel.topology,cgmodel.system,cgmodel.positions)
        >>> replica_energies,replica_positions,replica_state_indices = read_replica_exchange_data(system=cgmodel.system,topology=cgmodel.topology,temperature_list=,output_data="output.nc",print_frequency=None)


        """
        # Read the simulation coordinates for individual temperature replicas
        reporter = MultiStateReporter(output_data, open_mode='r')
        analyzer = ReplicaExchangeAnalyzer(reporter)

        replica_energies,unsampled_state_energies,neighborhoods,replica_state_indices = analyzer.read_energies()

        temps = np.array([temp._value for temp in temperature_list])
        beta_k = 1 / (kB * temps)
        for k in range(len(replica_energies)):
          replica_energies[:,k,:]*=beta_k[k]**(-1)

        total_steps = len(replica_energies[0][0])
        replica_positions = unit.Quantity(np.zeros([len(temperature_list),total_steps,system.getNumParticles(),3]),unit.nanometer)

        for step in range(total_steps):
          sampler_states = reporter.read_sampler_states(iteration=step)
          if type(sampler_states) == None:
            print("ERROR: no data found for step "+str(step))
            exit()
          else:
           for replica_index in range(len(temperature_list)):
             for particle in range(system.getNumParticles()):
                   replica_positions[replica_index][step][particle] = sampler_states[replica_index].positions[particle]

        return(replica_energies,replica_positions,replica_state_indices)

def run_replica_exchange(topology,system,positions,temperature_list=None,simulation_time_step=None,total_simulation_time=1.0 * unit.picosecond,output_data='output.nc',print_frequency=100,verbose_simulation=False,exchange_attempts=None,test_time_step=False,output_directory=None):
        """
        Run a Yank replica exchange simulation using an OpenMM coarse grained model.

        :param topology: OpenMM Topology
        :type topology: `Topology() <https://simtk.org/api_docs/openmm/api4_1/python/classsimtk_1_1openmm_1_1app_1_1topology_1_1Topology.html>`_

        :param system: OpenMM System()
        :type system: `System() <https://simtk.org/api_docs/openmm/api4_1/python/classsimtk_1_1openmm_1_1openmm_1_1System.html>`_

        :param positions: Positions array for the model we would like to test
        :type positions: `Quantity() <http://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_ ( np.array( [cgmodel.num_beads,3] ), simtk.unit )

        :param temperature_list: List of temperatures for which to perform replica exchange simulations, default = None
        :type temperature: List( float * simtk.unit.temperature )

        :param simulation_time_step: Simulation integration time step
        :type simulation_time_step: `SIMTK <https://simtk.org/>`_ `Unit() <http://docs.openmm.org/7.1.0/api-python/generated/simtk.unit.unit.Unit.html>`_

        :param total_simulation_time: Total run time for individual simulations
        :type total_simulation_time: `SIMTK <https://simtk.org/>`_ `Unit() <http://docs.openmm.org/7.1.0/api-python/generated/simtk.unit.unit.Unit.html>`_

        :param output_data: Name of NETCDF file where we will write simulation data
        :type output_data: string

        :param print_frequency: Number of simulation steps to skip when writing to output, Default = 100
        :type print_frequence: int

        :param verbose_simulation: Determines how much output is printed during a simulation run.  Default = False
        :type verbose_simulation: Logical

        :param exchange_attempts: Number of exchange attempts to make during a replica exchange simulation run, Default = None
        :type exchange_attempts: int

        :param test_time_step: Logical variable determining if a test of the time step will be performed, Default = False
        :type test_time_step: Logical

        :param output_directory: Path to which we will write the output from simulation runs.
        :type output_directory: str

        :returns: 
            - replica_energies ( `Quantity() <http://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_ ( np.float( [number_replicas,number_simulation_steps] ), simtk.unit ) ) - The potential energies for all replicas at all (printed) time steps
            - replica_positions ( `Quantity() <http://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_ ( np.float( [number_replicas,number_simulation_steps,cgmodel.num_beads,3] ), simtk.unit ) ) - The positions for all replicas at all (printed) time steps
 
            - replica_state_indices ( np.int64( [number_replicas,number_simulation_steps] ), simtk.unit ) - The thermodynamic state assignments for all replicas at all (printed) time steps

        :Example:

        >>> from foldamers.cg_model.cgmodel import CGModel
        >>> from cg_openmm.simulation.rep_exch import *
        >>> cgmodel = CGModel()
        >>> replica_energies,replica_positions,replica_state_indices = run_replica_exchange(cgmodel.topology,cgmodel.system,cgmodel.positions)

        """
        if simulation_time_step == None:
          simulation_time_step,force_threshold = get_simulation_time_step(topology,system,positions,temperature_list[-1],total_simulation_time)

        simulation_steps = int(round(total_simulation_time.__div__(simulation_time_step)))

        if exchange_attempts == None:
          if simulation_steps > 10000:
            exchange_attempts = round(simulation_steps/1000)
          else:
            exchange_attempts = 10

        if temperature_list == None:
          temperature_list = [(300.0 * unit.kelvin).__add__(i * unit.kelvin) for i in range(-50,50,10)]

        num_replicas = len(temperature_list)
        sampler_states = list()
        thermodynamic_states = list()

        # Define thermodynamic states.
        box_vectors = system.getDefaultPeriodicBoxVectors()
        for temperature in temperature_list:
          thermodynamic_state = mmtools.states.ThermodynamicState(system=system, temperature=temperature)
          thermodynamic_states.append(thermodynamic_state)
          sampler_states.append(mmtools.states.SamplerState(positions,box_vectors=box_vectors))

        # Create and configure simulation object.
        move = mmtools.mcmc.LangevinDynamicsMove(timestep=simulation_time_step,collision_rate=5.0/unit.picosecond,n_steps=exchange_attempts, reassign_velocities=True)
        simulation = ReplicaExchangeSampler(mcmc_moves=move, number_of_iterations=exchange_attempts)

        if os.path.exists(output_data): os.remove(output_data)
        reporter = MultiStateReporter(output_data, checkpoint_interval=1)
        simulation.create(thermodynamic_states, sampler_states, reporter)
        config_root_logger(verbose_simulation)
       
        if not test_time_step:
           num_attempts = 0
           while num_attempts < 5:
            try:
              simulation.run()
              #print("Replica exchange simulations succeeded with a time step of: "+str(simulation_time_step))
              break
            except:
              num_attempts = num_attempts + 1
           if num_attempts >= 5:
             print("Replica exchange simulation attempts failed, try verifying your model/simulation settings.")
             exit()
        else:
          simulation_time_step,force_threshold = get_simulation_time_step(topology,system,positions,temperature_list[-1],total_simulation_time)
          print("The suggested time step for a simulation with this model is: "+str(simulation_time_step))
          while simulation_time_step.__div__(2.0) > 0.001 * unit.femtosecond:
          
            try:
              print("Running replica exchange simulations with Yank...")
              print("Using a time step of "+str(simulation_time_step))
              print("Running each trial simulation for 1000 steps, with 10 exchange attempts.")
              move = mmtools.mcmc.LangevinDynamicsMove(timestep=simulation_time_step,collision_rate=20.0/unit.picosecond,n_steps=10, reassign_velocities=True)
              simulation = ReplicaExchangeSampler(replica_mixing_scheme='swap-neighbors',mcmc_moves=move,number_of_iterations=10)
              reporter = MultiStateReporter(output_data, checkpoint_interval=1)
              simulation.create(thermodynamic_states, sampler_states, reporter)

              simulation.run()
              print("Replica exchange simulations succeeded with a time step of: "+str(simulation_time_step))
              break
            except:
              del simulation
              os.remove(output_data)
              print("Simulation attempt failed with a time step of: "+str(simulation_time_step))
              if simulation_time_step.__div__(2.0) > 0.001 * unit.femtosecond:
                simulation_time_step = simulation_time_step.__div__(2.0)
              else:
                print("Error: replica exchange simulation attempt failed with a time step of: "+str(simulation_time_step))
                print("Please check the model and simulations settings, and try again.")
                exit()

        replica_energies,replica_positions,replica_state_indices = read_replica_exchange_data(system=system,topology=topology,temperature_list=temperature_list,output_data=output_data,print_frequency=print_frequency)

        steps_per_stage = round(simulation_steps/exchange_attempts)
        if output_directory != None:
          plot_replica_exchange_energies(replica_energies,temperature_list,simulation_time_step,steps_per_stage=steps_per_stage,output_directory=output_directory)
          plot_replica_exchange_summary(replica_state_indices,temperature_list,simulation_time_step,steps_per_stage=steps_per_stage,output_directory=output_directory)
        else:
          plot_replica_exchange_energies(replica_energies,temperature_list,simulation_time_step,steps_per_stage=steps_per_stage)
          plot_replica_exchange_summary(replica_state_indices,temperature_list,simulation_time_step,steps_per_stage=steps_per_stage)

        return(replica_energies,replica_positions,replica_state_indices)

def get_minimum_energy_ensemble(topology,replica_energies,replica_positions,ensemble_size=5,file_name=None):
        """
        Get an ensemble of low (potential) energy poses, and write the lowest energy structure to a PDB file if a file_name is provided.

        :param topology: OpenMM Topology()
        :type topology: `Topology() <https://simtk.org/api_docs/openmm/api4_1/python/classsimtk_1_1openmm_1_1app_1_1topology_1_1Topology.html>`_

        :param replica_energies: List of dimension num_replicas X simulation_steps, which gives the energies for all replicas at all simulation steps 
        :type replica_energies: List( List( float * simtk.unit.energy for simulation_steps ) for num_replicas )

        :param replica_positions: List of positions for all output frames for all replicas
        :type replica_positions: np.array( ( float * simtk.unit.positions for num_beads ) for simulation_steps )

        :param file_name: Output destination for PDB coordinates of minimum energy pose, Default = None

        :returns:
             - ensemble ( List() ) - A list of poses that are in the minimum energy ensemble.

        :Example:

        >>> from foldamers.cg_model.cgmodel import CGModel
        >>> from cg_openmm.simulation.rep_exch import *
        >>> cgmodel = CGModel()
        >>> replica_energies,replica_positions,replica_state_indices = run_replica_exchange(cgmodel.topology,cgmodel.system,cgmodel.positions)
        >>> ensemble_size = 5
        >>> file_name = "minimum.pdb"
        >>> minimum_energy_ensemble = get_minimum_energy_ensemble(cgmodel.topology,replica_energies,replica_positions,ensemble_size=ensemble_size,file_name=file_name)

        """
        # Get the minimum energy structure sampled during the simulation
        ensemble = []
        ensemble_energies = []
        for replica in range(len(replica_energies)):
          energies = np.array([energy for energy in replica_energies[replica][replica]])
          for energy in range(len(energies)):
            if len(ensemble) < ensemble_size:
              ensemble.append(replica_positions[replica][energy])
              ensemble_energies.append(energies[energy])
            else:

             for comparison in range(len(ensemble_energies)):
              if energies[energy] < ensemble_energies[comparison]:
               ensemble_energies[comparison] = energies[energy]
               ensemble[comparison] = replica_positions[replica][energy]

        if file_name == None:
         index = 1
         for pose in ensemble:
          file = open(str("re_min_"+str(index)+".pdb"),"w")
          PDBFile.writeFile(topology,pose,file=file)
        else:
          file = open(file_name,"w")
          for pose in ensemble:
            PDBFile.writeFile(topology,pose,file=file)

        return(ensemble)

def plot_replica_exchange_energies(replica_energies,temperature_list,simulation_time_step,steps_per_stage=1,file_name="rep_ex_ener.png",legend=True,output_directory=None):
        """
        Plot the potential energies for a batch of replica exchange trajectories

        :param replica_energies: List of dimension num_replicas X simulation_steps, which gives the energies for all replicas at all simulation steps 
        :type replica_energies: List( List( float * simtk.unit.energy for simulation_steps ) for num_replicas )

        :param temperature_list: List of temperatures for which to perform replica exchange simulations, default = [(300.0 * unit.kelvin).__add__(i * unit.kelvin) for i in range(-20,100,10)] 
        :type temperature: List( float * simtk.unit.temperature )

        :param simulation_time_step: Simulation integration time step
        :type simulation_time_step: `SIMTK <https://simtk.org/>`_ `Unit() <http://docs.openmm.org/7.1.0/api-python/generated/simtk.unit.unit.Unit.html>`_

        :param steps_per_stage: The number of simulation steps for individual replica "stages" (period of time between state exchanges), default = 1
        :type steps_per_stage: int

        :param file_name: The pathname of the output file for plotting results, default = "replica_exchange_energies.png"
        :type file_name: str

        :param output_directory: Path to which we will write the output from simulation runs, Default = None
        :type output_directory: str

        :param legend: Controls whether a legend is added to the plot
        :type legend: Logical

        ..warning:: If more than 10 replica exchange trajectories are provided as input data, by default, this function will only plot the first 10 thermodynamic states.  These thermodynamic states are chosen based upon their indices, not their instantaneous temperature (ensemble) assignment.

        """

        figure = pyplot.figure(0)

        for replica in range(len(replica_energies)):
          simulation_times = np.array([float(int(step*steps_per_stage)*simulation_time_step.in_units_of(unit.picosecond)._value) for step in range(len(replica_energies[replica][replica]))])
          energies = np.array([float(energy) for energy in replica_energies[replica][replica]])
          pyplot.plot(simulation_times,energies,figure=figure)

        pyplot.xlabel("Simulation Time ( Picoseconds )")
        pyplot.ylabel("Potential Energy ( kJ / mol )")
        pyplot.title("Replica Exchange Simulation")
        if legend:
         if len(temperature_list) > 10:
          pyplot.legend([round(temperature._value,1) for temperature in temperature_list[0:9]],loc='center left', bbox_to_anchor=(1, 0.5),title='T (K)')
         else:
          pyplot.legend([round(temperature._value,1) for temperature in temperature_list],loc='center left', bbox_to_anchor=(1, 0.5),title='T (K)')
        if output_directory != None:
          output_file = str(str(output_directory)+"/"+str(file_name))
          pyplot.savefig(output_file,bbox_inches='tight')
        else:
          pyplot.savefig(file_name,bbox_inches='tight')
        #pyplot.show()
        pyplot.close()

        return

def plot_replica_exchange_summary(replica_states,temperature_list,simulation_time_step,steps_per_stage=1,file_name="rep_ex_states.png",legend=True,output_directory=None):
        """
        Plot the thermodynamic state assignments for individual temperature replicas as a function of the simulation time, in order to obtain a visual summary of the replica exchanges from a Yank simulation.

        :param replica_states: List of dimension num_replicas X simulation_steps, which gives the thermodynamic state indices for all replicas at all simulation steps 
        :type replica_states: List( List( float * simtk.unit.energy for simulation_steps ) for num_replicas )

        :param temperature_list: List of temperatures for which to perform replica exchange simulations, default = [(300.0 * unit.kelvin).__add__(i * unit.kelvin) for i in range(-20,100,10)] 
        :type temperature: List( float * simtk.unit.temperature )

        :param simulation_time_step: Simulation integration time step
        :type simulation_time_step: `SIMTK <https://simtk.org/>`_ `Unit() <http://docs.openmm.org/7.1.0/api-python/generated/simtk.unit.unit.Unit.html>`_

        :param steps_per_stage: The number of simulation steps for individual replica "stages" (period of time between state exchanges), default = 1
        :type steps_per_stage: int

        :param file_name: The pathname of the output file for plotting results, default = "replica_exchange_state_transitions.png"
        :type file_name: str

        :param legend: Controls whether a legend is added to the plot
        :type legend: Logical

        :param output_directory: Path to which we will write the output from simulation runs, default = None
        :type output_directory: str

        ..warning:: If more than 10 replica exchange trajectories are provided as input data, by default, this function will only plot the first 10 thermodynamic states.  These thermodynamic states are chosen based upon their indices, not their instantaneous temperature (ensemble) assignment.

        """

        figure = pyplot.figure(1)
        for replica in range(len(replica_states)):
          simulation_times = np.array([float(int(step*steps_per_stage)*simulation_time_step.in_units_of(unit.picosecond)._value) for step in range(len(replica_states[replica]))])
          state_indices = np.array([int(round(state)) for state in replica_states[replica]])
          pyplot.plot(simulation_times,state_indices,figure=figure)

        pyplot.xlabel("Simulation Time ( Picoseconds )")
        pyplot.ylabel("Thermodynamic State Index")
        pyplot.title("State Exchange Summary")
        if legend:
         if len(replica_states) > 10:
          pyplot.legend([i for i in range(len(replica_states[0:9]))],loc='center left', bbox_to_anchor=(1, 0.5),title='Replica Index')
         else:
          pyplot.legend([i for i in range(len(replica_states))],loc='center left', bbox_to_anchor=(1, 0.5),title='Replica Index')
        if output_directory != None:
          output_file = str(str(output_directory)+"/"+str(file_name))
          pyplot.savefig(output_file,bbox_inches='tight')
        else:
          pyplot.savefig(file_name,bbox_inches='tight')
        #pyplot.show()
        pyplot.close()

        return



import numpy as np
from simtk import openmm as mm
from simtk.openmm import *
from simtk.openmm.vec3 import Vec3
from simtk import unit
from simtk.openmm.app.pdbfile import PDBFile
import mdtraj
import simtk.openmm.app.element as elem
from simtk.openmm.app import *
import matplotlib.pyplot as pyplot
import csv
from cg_openmm.utilities.iotools import write_bonds

def get_simulation_time_step(topology,system,positions,temperature,total_simulation_time,time_step_list=None):
        """
        Determine a suitable simulation time step.

        :param topology: OpenMM Topology
        :type topology: `Topology() <https://simtk.org/api_docs/openmm/api4_1/python/classsimtk_1_1openmm_1_1app_1_1topology_1_1Topology.html>`_

        :param system: OpenMM System()
        :type system: `System() <https://simtk.org/api_docs/openmm/api4_1/python/classsimtk_1_1openmm_1_1openmm_1_1System.html>`_

        :param positions: Positions array for the model we would like to test
        :type positions: `Quantity() <http://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_ ( np.array( [cgmodel.num_beads,3] ), simtk.unit )

        :param temperature: Simulation temperature
        :type temperature: `SIMTK <https://simtk.org/>`_ `Unit() <http://docs.openmm.org/7.1.0/api-python/generated/simtk.unit.unit.Unit.html>`_

        :param total_simulation_time: Total run time for individual simulations
        :type total_simulation_time: `SIMTK <https://simtk.org/>`_ `Unit() <http://docs.openmm.org/7.1.0/api-python/generated/simtk.unit.unit.Unit.html>`_

        :param time_step_list: List of time steps for which to attempt a simulation in OpenMM.
        :type time_step_list: List, default = None

        :returns: 
             - time_step ( `SIMTK <https://simtk.org/>`_ `Unit() <http://docs.openmm.org/7.1.0/api-python/generated/simtk.unit.unit.Unit.html>`_ ) - A successfully-tested simulation time-step for the provided coarse grained model
             - tolerance ( `SIMTK <https://simtk.org/>`_ `Unit() <http://docs.openmm.org/7.1.0/api-python/generated/simtk.unit.unit.Unit.html>`_ ) - The maximum change in forces that will be tolerated when testing the time step.

        :Example:

        >>> from simtk import unit
        >>> from foldamers.cg_model.cgmodel import CGModel
        >>> cgmodel = CGModel()
        >>> topology = cgmodel.topology
        >>> system = cgmodel.system
        >>> positions = cgmodel.positions
        >>> temperature = 300.0 * unit.kelvin
        >>> total_simulation_time = 1.0 * unit.picosecond
        >>> time_step_list = [1.0 * unit.femtosecond, 2.0 * unit.femtosecond, 5.0 * unit.femtosecond]
        >>> best_time_step,max_force_tolerance = get_simulation_time_step(topology,system,positions,temperature,total_simulation_time,time_step_list=time_step_list)

        """
        tolerance = 10.0
        success = False

        if time_step_list == None:
          time_step_list = [10.0-i * unit.femtosecond for i in [5.0,7.5,9.0,9.5,9.9,9.99]]

        if type(time_step_list) != list:
          time_step_list = [time_step_list]
        for time_step in time_step_list:
          integrator = LangevinIntegrator(temperature._value,1.0 / unit.picoseconds, time_step.in_units_of(unit.picosecond)._value)

          simulation = Simulation(topology, system, integrator)
          simulation.context.setPositions(positions.in_units_of(unit.nanometer))
          simulation.context.setVelocitiesToTemperature(temperature)
          simulation.reporters.append(PDBReporter('test.pdb',1))
          simulation.reporters.append(StateDataReporter('test.dat',1, \
          step=True, totalEnergy=True, potentialEnergy=True, kineticEnergy=True, temperature=True))

          total_steps = round(total_simulation_time.__div__(time_step))
          try:
            simulation.minimizeEnergy()
            positions = simulation.context.getState(getPositions=True).getPositions()
            success = True
            break
          except:
            continue
        if not success:
          for tolerance in [10 ** exponent for exponent in range(2,10)]:
            try:
#            simulation.context.applyConstraints(1.0e-8)
             integrator = LangevinIntegrator(temperature._value,1.0 / unit.picoseconds, time_step.in_units_of(unit.picosecond)._value)
             simulation = Simulation(topology, system, integrator)
             simulation.context.setPositions(positions)
             simulation.minimizeEnergy(tolerance=tolerance)
             success = True
             break
            except:
             continue
        if not success:
          tolerance = None
        return(time_step,tolerance)

def minimize_structure(topology,system,positions,temperature=0.0 * unit.kelvin,simulation_time_step=None,total_simulation_time=1.0 * unit.picosecond,output_pdb=None,output_data=None,print_frequency=1):
        """
        Minimize the potential energy

        :param topology: OpenMM topology
        :type topology: Topology()

        :param system: OpenMM system
        :type system: System()

        :param positions: Positions array for the model we would like to test
        :type positions: `Quantity() <http://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_ ( np.array( [cgmodel.num_beads,3] ), simtk.unit )

        :param temperature: Simulation temperature
        :type temperature: `SIMTK <https://simtk.org/>`_ `Unit() <http://docs.openmm.org/7.1.0/api-python/generated/simtk.unit.unit.Unit.html>`_

        :param total_simulation_time: Total run time for individual simulations
        :type total_simulation_time: `SIMTK <https://simtk.org/>`_ `Unit() <http://docs.openmm.org/7.1.0/api-python/generated/simtk.unit.unit.Unit.html>`_

        :param output_pdb: Output destinaton for PDB-formatted coordinates during the simulation
        :type output_pdb: str

        :param output_data: Output destination for simulation data
        :type output_data: str

        :param print_frequency: Number of simulation steps to skip when writing data, default = 1
        :type print_frequency: int

        :returns: 
             - positions ( `Quantity() <http://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_ ( np.array( [cgmodel.num_beads,3] ), simtk.unit ) ) - Minimized positions
             - potential_energy ( `Quantity() <http://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_ - Potential energy for the minimized structure. 

        :Example:

        >>> from simtk import unit
        >>> from foldamers.cg_model.cgmodel import CGModel
        >>> cgmodel = CGModel()
        >>> topology = cgmodel.topology
        >>> system = cgmodel.system
        >>> positions = cgmodel.positions
        >>> temperature = 300.0 * unit.kelvin
        >>> total_simulation_time = 1.0 * unit.picosecond
        >>> simulation_time_step = 1.0 * unit.femtosecond
        >>> output_pdb = "output.pdb"
        >>> output_data = "output.dat"
        >>> print_frequency = 20
        >>> minimum_energy_structure,potential_energy,openmm_simulation_object = minimize_structure(topology,system,positions,temperature=temperature,simulation_time_step=simulation_time_step,total_simulation_time=total_simulation_time,output_pdb=output_pdb,output_data=output_data,print_frequency=print_frequency)

        """
        if simulation_time_step == None:
          print("Minimizing the structure, but no time step was provided.")
          exit()
          simulation_time_step_list = [(10.0 * (0.5 ** i)) * unit.femtosecond for i in range(0,14)]
          time_step,tolerance = get_simulation_time_step(topology,system,positions,temperature,total_simulation_time,simulation_time_step_list)
          if tolerance == None:
#            print("This set of positions is not a reasonable initial configuration.")
            energy = "NaN"
            simulation = None
            return(positions,energy)
        else:
          time_step = simulation_time_step
        integrator = LangevinIntegrator(temperature._value,1.0 / unit.picoseconds, time_step.in_units_of(unit.picosecond)._value)


        simulation = Simulation(topology, system, integrator)
        simulation.context.setPositions(positions.in_units_of(unit.nanometer))
        simulation.context.setVelocitiesToTemperature(temperature)
        forces = simulation.context.getState(getForces=True).getForces()
        if output_pdb != None:
          simulation.reporters.append(PDBReporter(output_pdb,print_frequency))
        if output_data != None:
          simulation.reporters.append(StateDataReporter(output_data,print_frequency, \
          step=True, totalEnergy=True, potentialEnergy=True, kineticEnergy=True, temperature=True))


        total_steps = round(total_simulation_time.__div__(time_step))
        potential_energy = None
        try:
          simulation.minimizeEnergy() # Set the simulation type to energy minimization
          positions = simulation.context.getState(getPositions=True).getPositions()
          potential_energy = simulation.context.getState(getEnergy=True).getPotentialEnergy()
        except:
          print("Minimization attempt failed with a time step of: "+str(time_step))
          if time_step.__gt__(0.01 * unit.femtosecond):
            time_step = time_step / 2.0
            print("Attempting minimization with a smaller time step.")
            positions,potential_energy = minimize_structure(topology,system,positions,temperature=temperature,simulation_time_step=time_step,total_simulation_time=total_simulation_time,output_pdb=output_pdb,output_data=output_data,print_frequency=print_frequency)
            time_step = time_step / 2.0 
          if time_step.__le__(0.01 * unit.femtosecond):
            print("Try using the 'get_simulation_time_step()' function,")
            print("or changing the 'simulation_time_step',")
            print("to see if one of these changes solves the problem.")
            #exit()

        return(positions,potential_energy,simulation)

def get_mm_energy(topology,system,positions):
        """
        Get the OpenMM potential energy for a system, given a topology and set of positions.

        :param topology: OpenMM Topology()
        :type topology: `Topology() <https://simtk.org/api_docs/openmm/api4_1/python/classsimtk_1_1openmm_1_1app_1_1topology_1_1Topology.html>`_

        :param system: OpenMM System()
        :type system: `System() <https://simtk.org/api_docs/openmm/api4_1/python/classsimtk_1_1openmm_1_1openmm_1_1System.html>`_

        :param positions: Positions array for the model we would like to test
        :type positions: `Quantity() <http://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_ ( np.array( [cgmodel.num_beads,3] ), simtk.unit )

        :returns: 
             - potential_energy ( `Quantity() <http://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_ ) - The potential energy for the model with the provided positions.
 
        :Example:

        >>> from foldamers.cg_model.cgmodel import CGModel
        >>> cgmodel = CGModel()
        >>> topology = cgmodel.topology
        >>> system = cgmodel.system
        >>> positions = cgmodel.positions
        >>> openmm_potential_energy = get_mm_energy(topology,system,positions)

        """
        simulation_time_step = 5.0 * unit.femtosecond
        friction = 0.0 / unit.picosecond
        integrator = LangevinIntegrator(0.0 * unit.kelvin,friction,simulation_time_step.in_units_of(unit.picosecond))
        simulation = Simulation(topology, system, integrator)
        simulation.context.setPositions(positions)
        potential_energy = simulation.context.getState(getEnergy=True).getPotentialEnergy()

        return(potential_energy)

def build_mm_simulation(topology,system,positions,temperature=300.0 * unit.kelvin,simulation_time_step=None,total_simulation_time=1.0 * unit.picosecond,output_pdb=None,output_data=None,print_frequency=100,test_time_step=False):
        """
        Build an OpenMM Simulation()

        :param topology: OpenMM Topology()
        :type topology: `Topology() <https://simtk.org/api_docs/openmm/api4_1/python/classsimtk_1_1openmm_1_1app_1_1topology_1_1Topology.html>`_

        :param system: OpenMM System()
        :type system: `System() <https://simtk.org/api_docs/openmm/api4_1/python/classsimtk_1_1openmm_1_1openmm_1_1System.html>`_

        :param positions: Positions array for the model we would like to test
        :type positions: `Quantity() <https://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_ ( np.array( [cgmodel.num_beads,3] ), simtk.unit )

        :param temperature: Simulation temperature, default = 300.0 K
        :type temperature: `SIMTK <https://simtk.org/>`_ `Unit() <http://docs.openmm.org/7.1.0/api-python/generated/simtk.unit.unit.Unit.html>`_

        :param simulation_time_step: Simulation integration time step
        :type simulation_time_step: `SIMTK <https://simtk.org/>`_ `Unit() <http://docs.openmm.org/7.1.0/api-python/generated/simtk.unit.unit.Unit.html>`_

        :param total_simulation_time: Total run time for individual simulations
        :type total_simulation_time: `SIMTK <https://simtk.org/>`_ `Unit() <http://docs.openmm.org/7.1.0/api-python/generated/simtk.unit.unit.Unit.html>`_

        :param output_pdb: Output destination for PDB coordinates, Default = None
        :type output_pdb: str

        :param output_data: Output destination for non-coordinate simulation data, Default = None
        :type output_data: str

        :param print_frequency: Number of simulation steps to skip when writing to output, Default = 100
        :type print_frequence: int
 
        :param test_time_step: Logical variable determining if a test of the time step will be performed, Default = False
        :type test_time_step: Logical

        :returns: 
             - simulation ( `Simulation() <https://simtk.org/api_docs/openmm/api4_1/python/classsimtk_1_1openmm_1_1app_1_1simulation_1_1Simulation.html>`_ ) - OpenMM Simulation() object

        :Example:

        >>> from simtk import unit
        >>> from foldamers.cg_model.cgmodel import CGModel
        >>> cgmodel = CGModel()
        >>> topology = cgmodel.topology
        >>> system = cgmodel.system
        >>> positions = cgmodel.positions
        >>> temperature = 300.0 * unit.kelvin
        >>> simulation_time_step = 5.0 * unit.femtosecond
        >>> total_simulation_time= 1.0 * unit.picosecond
        >>> output_pdb = "output.pdb"
        >>> output_data = "output.dat"
        >>> print_frequency = 20
        >>> openmm_simulation = build_mm_simulation(topology,system,positions,temperature=temperature,simulation_time_step=simulation_time_step,total_simulation_time=total_simulation_time,output_pdb=output_pdb,output_data=output_data,print_frequency=print_frequency,test_time_step=False)

        """
        if simulation_time_step == None:
#          print("No simulation time step provided.")
#          print("Going to attempt a range of time steps,")
#          print("to confirm their validity for these model settings,")
#          print("before performing a full simulation.")
          time_step_list = [(10.0 * (0.5 ** i)) * unit.femtosecond for i in range(0,14)]
          simulation_time_step,force_cutoff = get_simulation_time_step(topology,system,positions,temperature,total_simulation_time,time_step_list)
        friction = 1.0 / unit.picosecond

        integrator = LangevinIntegrator(temperature._value,friction,simulation_time_step.in_units_of(unit.picosecond)._value)
        
        #file = open("temp.pdb","w")
        #PDBFile.writeFile(topology,positions,file=file)
        #file.close()
        #pdb = PDBFile("temp.pdb")
        simulation = Simulation(topology, system, integrator)
        #os.remove("temp.pdb")
        #print("Setting positions")
        #print(type(positions))
        #positions = [Vec3(c[0].in_units_of(unit.nanometer)._value,c[1].in_units_of(unit.nanometer)._value,c[2].in_units_of(unit.nanometer)._value) for c in positions]
        simulation.context.setPositions(positions)
#        simulation.context.setVelocitiesToTemperature(temperature)
        if output_pdb != None:
          simulation.reporters.append(PDBReporter(output_pdb,print_frequency))
        if output_data != None:
          simulation.reporters.append(StateDataReporter(output_data,print_frequency, \
          step=True, totalEnergy=True, potentialEnergy=True, kineticEnergy=True, temperature=True))

#        simulation.minimizeEnergy() # Set the simulation type to energy minimization

        if test_time_step:
          try:
            simulation_temp = simulation.__deepcopy__(memo={})
            simulation_temp.step(100)
            print("Simulation attempt successful.")
          except:
#            print("Simulation attempt failed with a time step of: "+str(simulation_time_step))
#            print("Going to attempt to identify a smaller time step that allows simulation for this model and its current settings...")
            time_step_list = [(10.0 * (0.5 ** i)) * unit.femtosecond for i in range(0,14)]
            if all(simulation_time_step.__lt__(time_step) for time_step in time_step_list):
              print("Error: couldn't identify a suitable simulation time step for this model.")
              print("Check the model settings, consider changing the input time step,")
              print("and if this doesn't fix the problem, try changing the default list of time steps")
              print("that are sampled in .build.cg_build.build_mm_simulation.py'")
              exit()
            for time_step in time_step_list:
              if time_step < simulation_time_step:
                simulation = build_mm_simulation(topology,system,positions,temperature=temperature,simulation_time_step=time_step,total_simulation_time=total_simulation_time,output_pdb=output_pdb,output_data=output_data,print_frequency=print_frequency)
                try:
                  
                  simulation_temp.step(100)
                  return(simulation)
                except:
                  continue
        return(simulation)


def run_simulation(cgmodel,output_directory,total_simulation_time,simulation_time_step,temperature,print_frequency,output_pdb=None,output_data=None):
        """

        Run OpenMM() simulation

        :param cgmodel: CGModel() object
        :type cgmodel: class

        :param output_directory: Output directory for simulation data
        :type output_directory: str

        :param total_simulation_time: Total run time for individual simulations
        :type total_simulation_time: `SIMTK <https://simtk.org/>`_ `Unit() <http://docs.openmm.org/7.1.0/api-python/generated/simtk.unit.unit.Unit.html>`_

        :param simulation_time_step: Simulation integration time step
        :type simulation_time_step: `SIMTK <https://simtk.org/>`_ `Unit() <http://docs.openmm.org/7.1.0/api-python/generated/simtk.unit.unit.Unit.html>`_

        :param temperature: Simulation temperature, default = 300.0 K
        :type temperature: `SIMTK <https://simtk.org/>`_ `Unit() <http://docs.openmm.org/7.1.0/api-python/generated/simtk.unit.unit.Unit.html>`_

        :param print_frequency: Number of simulation steps to skip when writing to output, Default = 100
        :type print_frequence: int

        :Example:

        >>> import os
        >>> from simtk import unit
        >>> from foldamers.cg_model.cgmodel import CGModel
        >>> cgmodel = CGModel()
        >>> topology = cgmodel.topology
        >>> system = cgmodel.system
        >>> positions = cgmodel.positions
        >>> temperature = 300.0 * unit.kelvin
        >>> simulation_time_step = 5.0 * unit.femtosecond
        >>> total_simulation_time= 1.0 * unit.picosecond
        >>> output_directory = os.getcwd()
        >>> output_pdb = "output.pdb"
        >>> output_data = "output.dat"
        >>> print_frequency = 20
        >>> run_simulation(cgmodel,output_directory,total_simulation_time,simulation_time_step,temperature,print_frequency,output_pdb=output_pdb,output_data=output_data)

        .. warning:: When run with default options this subroutine is capable of producing a large number of output files.  For example, by default this subroutine will plot the simulation data that is written to an output file.

        """
        total_steps = round(total_simulation_time.__div__(simulation_time_step))
        if not os.path.exists(output_directory): os.mkdir(output_directory)
        if output_pdb == None:
          output_pdb = str(str(output_directory)+'/simulation.pdb')
        else:
          output_pdb = str(str(output_directory)+"/"+str(output_pdb))
        if output_data == None:
          output_data = str(str(output_directory)+'/simulation.dat')
        else:
          output_data = str(str(output_directory)+"/"+str(output_data))

        simulation = build_mm_simulation(cgmodel.topology,cgmodel.system,cgmodel.positions,total_simulation_time=total_simulation_time,simulation_time_step=simulation_time_step,temperature=temperature,output_pdb=output_pdb,output_data=output_data,print_frequency=print_frequency)

        for step in range(total_steps):
          sim = simulation
          try:
            sim.step(1)
            simulation = sim
          except:
            attempts = 1
            while attempts <= 3:
              try:
                sim = simulation
                sim.step(1)
                simulation = sim
              except:
                attempts = attempts + 1
            if attempts > 3:
              plot_simulation_results(output_data,output_directory,simulation_time_step,total_simulation_time)
              print("Error: simulation attempt failed.")
              exit()

        if not cgmodel.include_bond_forces and cgmodel.constrain_bonds:
          file = open(output_pdb,"r")
          lines = file.readlines()
          file.close()
          os.remove(output_pdb)
          file = open(output_pdb,"w")
          for line in lines[:-1]:
            file.write(line)
          write_bonds(cgmodel,file)
          file.close()

        plot_simulation_results(output_data,output_directory,simulation_time_step,total_simulation_time)
        return

def read_simulation_data(simulation_data_file,simulation_time_step):
        """
        Read OpenMM simulation data

        :param simulation_data_file: Path to file that will be read
        :type simulation_data_file: str

        :param simulation_time_step: Time step to apply for the simulation data
        :type simulation_time_step: `SIMTK <https://simtk.org/>`_ `Unit() <http://docs.openmm.org/7.1.0/api-python/generated/simtk.unit.unit.Unit.html>`_

        :returns:
          - data ( dict( "Simulation Time": list,"Potential Energy": list,"Kinetic Energy": list,"Total Energy": list,"Temperature": list ) ) - A dictionary containing the simulation times, potential energies, kinetic energies, and total energies from an OpenMM simulation trajectory.

        :Example:

        >>> from simtk import unit
        >>> simulation_data_file = "output.dat"
        >>> simulation_time_step = 5.0 * unit.femtosecond
        >>> data = read_simulation_data(simulation_data_file,simulation_time_step)

        """
        data = {"Simulation Time":[],"Potential Energy":[],"Kinetic Energy":[],"Total Energy":[],"Temperature":[]}
        with open(simulation_data_file,newline='') as csvfile:
          reader = csv.reader(csvfile, delimiter=',')
          next(reader)
          for row in reader:
            data["Simulation Time"].append(float(simulation_time_step.in_units_of(unit.picosecond)._value)*float(row[0]))
            data["Potential Energy"].append(float(row[1]))
            data["Kinetic Energy"].append(float(row[2]))
            data["Total Energy"].append(float(row[3]))
            data["Temperature"].append(float(row[4]))

        return(data)

def plot_simulation_data(simulation_times,y_data,plot_type=None,output_directory=None):
        """
        Plot simulation data.

        :param simulation_times: List of simulation times (x data)
        :type simulation_times: List

        :param y_data: List of simulation data
        :type y_data: List

        :param plot_type: Form of data to plot, Default = None, Valid options include: "Potential Energy", "Kinetic Energy", "Total Energy", "Temperature"
        :type plot_type: str

        :Example:

        >>> import os
        >>> from simtk import unit
        >>> simulation_data_file = "output.pdb"
        >>> simulation_time_step = 5.0 * unit.femtosecond
        >>> simulation_data = read_simulation_data(simulation_data_file,simulation_time_step)
        >>> simulation_times = simulation_data["Simulation Time"]
        >>> y_data = simulation_data["Potential Energy"]
        >>> plot_type = "Potential Energy"
        >>> plot_simulation_data(simulation_times,y_data,plot_type="Potential Energy")

        """
        figure = pyplot.figure(1)
        pyplot.xlabel("Simulation Time (Picoseconds)")
        if plot_type == "Potential Energy":
          file_name = "Potential_Energy.png"
          pyplot.ylabel("Potential Energy (kJ/mole)")
          pyplot.title("Simulation Potential Energy")
        pyplot.xlabel("Simulation Time (Picoseconds)")
        if plot_type == "Kinetic Energy":
          file_name = "Kinetic_Energy.png"
          pyplot.ylabel("Kinetic Energy (kJ/mole)")
          pyplot.title("Simulation Kinetic Energy")
        pyplot.xlabel("Simulation Time (Picoseconds)")
        if plot_type == "Total Energy":
          file_name = "Total_Energy.png"
          pyplot.ylabel("Total Energy (kJ/mole)")
          pyplot.title("Simulation Total Energy")
        pyplot.xlabel("Simulation Time (Picoseconds)")
        if plot_type == "Temperature":
          file_name = "Temperature.png"
          pyplot.ylabel("Temperature (Kelvin)")
          pyplot.title("Simulation Temperature")

        pyplot.plot(simulation_times,y_data)
        if output_directory == None:
          pyplot.savefig(file_name)
        else:
          output_file = str(str(output_directory)+"/"+str(file_name))
          pyplot.savefig(output_file)
        pyplot.close()
        return

def plot_simulation_results(simulation_data_file,plot_output_directory,simulation_time_step,total_simulation_time):
        """
        Plot all data from an OpenMM output file

        :param simulation_data_file: Path to file containing simulation data
        :type simulation_data_file: str

        :param plot_output_directory: Path to folder where plotting results will be written.
        :type plot_output_directory: str

        :param simulation_time_step: Simulation integration time step
        :type simulation_time_step: `SIMTK <https://simtk.org/>`_ `Unit() <http://docs.openmm.org/7.1.0/api-python/generated/simtk.unit.unit.Unit.html>`_

        :Example:

        >>> import os
        >>> from simtk import unit
        >>> simulation_data_file = "output.pdb"
        >>> plot_output_directory = os.getcwd()
        >>> simulation_time_step = 5.0 * unit.femtosecond
        >>> plot_simulation_results(simulation_data_file,plot_output_directory,simulation_time_step)

        """
        data = read_simulation_data(simulation_data_file,simulation_time_step)
        
        plot_simulation_data(data["Simulation Time"],data["Potential Energy"],plot_type="Potential Energy",output_directory=plot_output_directory)
        plot_simulation_data(data["Simulation Time"],data["Kinetic Energy"],plot_type="Kinetic Energy",output_directory=plot_output_directory)
        plot_simulation_data(data["Simulation Time"],data["Total Energy"],plot_type="Total Energy",output_directory=plot_output_directory)
        plot_simulation_data(data["Simulation Time"],data["Temperature"],plot_type="Temperature",output_directory=plot_output_directory)
        return



import csv
from simtk import unit

def write_bonds(CGModel,pdb_object):
        """
        Writes the bonds from an input CGModel class object to the file object 'pdb_object', using PDB 'CONECT' syntax.

        :param CGModel: CGModel() class object
        :type CGModel: class

        :param pdb_object: File object to which we will write the bond list
        :type pdb_object: file

        """
        bond_list = CGModel.bond_list
        for bond in bond_list:
         if int(bond[0]) < int(bond[1]):
          pdb_object.write("CONECT"+str("{:>5}".format(bond[0]+1))+str("{:>5}".format(bond[1]+1))+"\n")      
         else:
          pdb_object.write("CONECT"+str("{:>5}".format(bond[1]+1))+str("{:>5}".format(bond[0]+1))+"\n")
        pdb_object.write(str("END\n"))
        return

def get_topology_from_pdbfile(filename):
        """
        Reads the positions from a PDB file.

        :param filename: Path to the file where we will read PDB coordinates.
        :type filename: str

        :returns:
          - topology ( `Topology() <https://simtk.org/api_docs/openmm/api4_1/python/classsimtk_1_1openmm_1_1app_1_1topology_1_1Topology.html>`_ ) - OpenMM Topology() object, which stores bonds, angles, and other structural attributes of the coarse grained model
        """
        pdb_mm_obj = PDBFile(filename)
        topology = pdb_mm_obj.getTopology()

        return(topology)

def get_positions_from_pdbfile(filename):
        """
        Reads the positions from a PDB file.

        :param filename: Path to the file where we will read PDB coordinates.
        :type filename: str

        :returns:
          - positions ( `Quantity() <https://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_ ( np.array( [cgmodel.num_beads,3] ), simtk.unit ) ) - Positions for the particles in the PDB file.
        """

        pdb_mm_obj = PDBFile(filename)
        positions = pdb_mm_obj.getPositions()

        return(positions)


def read_pdbfile(cgmodel,filename):
        """
        Reads the positions and topology from a PDB file.

        :param cgmodel: CGModel() class object
        :type cgmodel: class

        :param filename: Path to the file where we will read PDB coordinates.
        :type filename: str

        :returns:
          - cgmodel - A CGModel() class object with the positions and topology from the PDB file that was read.

        """

        pdb_mm_obj = PDBFile(filename)
        cgmodel.positions = pdb_mm_obj.getPositions()
        cgmodel.topology = pdb_mm_obj.getTopology()

        return(cgmodel)


def write_pdbfile_without_topology(CGModel,filename,energy=None):
        """
        Writes the positions from an input CGModel class object to the file 'filename'.

        :param CGModel: CGModel() class object
        :type CGModel: class

        :param filename: Path to the file where we will write PDB coordinates.
        :type filename: str

        :param energy: Energy to write to the PDB file, default = None
        :type energy: `Quantity() <https://docs.openmm.org/development/api-python/generated/simtk.unit.quantity.Quantity.html>`_

        """

        pdb_object = open(filename,"w")
        if energy != None:
          pdb_object.write("## The OpenMM potential energy for this structure is: "+str(energy)+"\n")
          pdb_object.write("## with the following parameter settings:\n")
          pdb_object.write("## sigma = "+str(CGModel.sigmas['bb_bb_sigma'])+", epsilon = "+str(CGModel.epsilons['bb_bb_eps'])+"\n")

        coordinates = CGModel.positions
        bead_index = 1
        for monomer_index in range(CGModel.polymer_length):
          monomer_type = CGModel.sequence[monomer_index]
          element_index = 1
          for backbone_bead in range(monomer_type['backbone_length']):

            if monomer_index in list([0,CGModel.polymer_length-1]):
             pdb_object.write(str("ATOM"+str("{:>7}".format(bead_index))+" X"+str(element_index)+str("{:>6}".format(str("MT")))+" A"+str("{:>4}".format(monomer_index+1))+"     "+str("{:>7}".format(round(coordinates[bead_index-1][0].in_units_of(unit.angstrom)._value,2)))+" "+str("{:>7}".format(round(coordinates[bead_index-1][1].in_units_of(unit.angstrom)._value,2)))+" "+str("{:>7}".format(round(coordinates[bead_index-1][2].in_units_of(unit.angstrom)._value,2)))+"  1.00  0.00\n"))
            else:
             pdb_object.write(str("ATOM"+str("{:>7}".format(bead_index))+" X"+str(element_index)+str("{:>6}".format(str("M")))+" A"+str("{:>4}".format(monomer_index+1))+"     "+str("{:>7}".format(round(coordinates[bead_index-1][0].in_units_of(unit.angstrom)._value,2)))+" "+str("{:>7}".format(round(coordinates[bead_index-1][1].in_units_of(unit.angstrom)._value,2)))+" "+str("{:>7}".format(round(coordinates[bead_index-1][2].in_units_of(unit.angstrom)._value,2)))+"  1.00  0.00\n"))
            bead_index = bead_index + 1
            element_index = element_index + 1

            if backbone_bead in [monomer_type['sidechain_positions']]:
              for sidechain_bead in range(monomer_type['sidechain_length']):
                if monomer_index in list([0,CGModel.polymer_length-1]):
                 pdb_object.write(str("ATOM"+str("{:>7}".format(bead_index))+" A"+str(element_index)+str("{:>6}".format(str("MT")))+" A"+str("{:>4}".format(monomer_index+1))+"     "+str("{:>7}".format(round(coordinates[bead_index-1][0].in_units_of(unit.angstrom)._value,2)))+" "+str("{:>7}".format(round(coordinates[bead_index-1][1].in_units_of(unit.angstrom)._value,2)))+" "+str("{:>7}".format(round(coordinates[bead_index-1][2].in_units_of(unit.angstrom)._value,2)))+"  1.00  0.00\n"))
                else:
                 pdb_object.write(str("ATOM"+str("{:>7}".format(bead_index))+" A"+str(element_index)+str("{:>6}".format(str("M")))+" A"+str("{:>4}".format(monomer_index+1))+"     "+str("{:>7}".format(round(coordinates[bead_index-1][0].in_units_of(unit.angstrom)._value,2)))+" "+str("{:>7}".format(round(coordinates[bead_index-1][1].in_units_of(unit.angstrom)._value,2)))+" "+str("{:>7}".format(round(coordinates[bead_index-1][2].in_units_of(unit.angstrom)._value,2)))+"  1.00  0.00\n"))
                bead_index = bead_index + 1
                element_index = element_index + 1
        pdb_object.write(str("TER\n"))

        write_bonds(CGModel,pdb_object)
        pdb_object.close()
        return

def read_mm_energies(openmm_data_file):
          """
          Read the energies from an OpenMM data file.

          :param openmm_data_file: The path to an OpenMM data file (CSV format)
          :type openmm_data_file: str

          :returns: 
              - energies ( np.array( float * simtk.unit ) ) - An array containing all data in 'openmm_data_file'

          """

          energies = {'step': [],'potential_energy': [],'kinetic_energy': [], 'total_energy': [], 'temperature': []}

          with open(openmm_data_file) as csvfile:
            readCSV = csv.DictReader(csvfile,delimiter=',')
            for row in readCSV:
              try:
                energies['step'].append(row['#"Step"'])
              except:
                continue
              try:
                energies['potential_energy'].append(row['Potential Energy (kJ/mole)'])
              except:
                continue
              try:
                energies['kinetic_energy'].append(row['Kinetic Energy (kJ/mole)'])
              except:
                continue
              try:
                energies['total_energy'].append(row['Total Energy (kJ/mole)'])
              except:
                continue
              try:
                energies['temperature'].append(row['Temperature (K)'])
              except:
                continue

#              print(energies)
#              exit()

          return(energies)

